import pytest

from tests.base import DBTestCase, engine_version_lt, is_running_sqlite
from tests.example_apps.music.tables import Band, Manager


class TestInsert(DBTestCase):
    def test_insert(self):
        self.insert_rows()

        Band.insert(Band(name="Rustaceans", popularity=100)).run_sync()

        response = Band.select(Band.name).run_sync()
        names = [i["name"] for i in response]

        self.assertIn("Rustaceans", names)

    def test_add(self):
        self.insert_rows()

        Band.insert().add(Band(name="Rustaceans", popularity=100)).run_sync()

        response = Band.select(Band.name).run_sync()
        names = [i["name"] for i in response]

        self.assertIn("Rustaceans", names)

    def test_incompatible_type(self):
        """
        You shouldn't be able to add instances of a different table.
        """
        with self.assertRaises(TypeError):
            Band.insert().add(Manager(name="Guido")).run_sync()

    def test_insert_curly_braces(self):
        """
        You should be able to insert curly braces without an error.
        """
        self.insert_rows()

        Band.insert(Band(name="{}", popularity=100)).run_sync()

        response = Band.select(Band.name).run_sync()
        names = [i["name"] for i in response]

        self.assertIn("{}", names)

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_insert_returning(self):
        """
        Make sure update works with the `returning` clause.
        """
        response = (
            Manager.insert(Manager(name="Maz"))
            .returning(Manager.name)
            .run_sync()
        )

        self.assertListEqual(response, [{"name": "Maz"}])

    @pytest.mark.skipif(
        is_running_sqlite() and engine_version_lt(3.35),
        reason="SQLite version not supported",
    )
    def test_insert_returning_alias(self):
        """
        Make sure update works with the `returning` clause.
        """
        response = (
            Manager.insert(Manager(name="Maz"))
            .returning(Manager.name.as_alias("manager_name"))
            .run_sync()
        )

        self.assertListEqual(response, [{"manager_name": "Maz"}])



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

from piccolo.apps.migrations.auto.operations import (
    AddColumn,
    AlterColumn,
    DropColumn,
)
from piccolo.apps.migrations.auto.serialisation import (
    deserialise_params,
    serialise_params,
)
from piccolo.columns.base import Column
from piccolo.table import Table, create_table_class


def compare_dicts(dict_1, dict_2) -> t.Dict[str, t.Any]:
    """
    Returns a new dictionary which only contains key, value pairs which are in
    the first dictionary and not the second.

    For example:
        dict_1 = {'a': 1, 'b': 2}
        dict_2 = {'a': 1}
        returns {'b': 2}

        dict_1 = {'a': 2, 'b': 2}
        dict_2 = {'a': 1}
        returns {'a': 2, 'b': 2}

    """
    output = {}

    for key, value in dict_1.items():

        dict_2_value = dict_2.get(key, ...)
        if (
            dict_2_value is not ...
            and dict_2_value != value
            or dict_2_value is ...
        ):
            output[key] = value

    return output


@dataclass
class TableDelta:
    add_columns: t.List[AddColumn] = field(default_factory=list)
    drop_columns: t.List[DropColumn] = field(default_factory=list)
    alter_columns: t.List[AlterColumn] = field(default_factory=list)

    def __eq__(self, value: TableDelta) -> bool:  # type: ignore
        """
        This is mostly for testing purposes.
        """
        return True


@dataclass
class ColumnComparison:
    """
    As Column overrides it's `__eq__` method, to allow it to be used in the
    `where` clause of a query, we need to wrap `Column` if we want to compare
    them.
    """

    column: Column

    def __hash__(self) -> int:
        return self.column.__hash__()

    def __eq__(self, value) -> bool:
        if isinstance(value, ColumnComparison):
            return (
                self.column._meta.db_column_name
                == value.column._meta.db_column_name
            )
        return False


@dataclass
class DiffableTable:
    """
    Represents a Table. When we substract two instances, it returns the
    changes.
    """

    class_name: str
    tablename: str
    columns: t.List[Column] = field(default_factory=list)
    previous_class_name: t.Optional[str] = None

    def __post_init__(self):
        self.columns_map: t.Dict[str, Column] = {
            i._meta.name: i for i in self.columns
        }

    def __sub__(self, value: DiffableTable) -> TableDelta:
        if not isinstance(value, DiffableTable):
            raise ValueError(
                "Can only diff with other DiffableTable instances"
            )

        if value.class_name != self.class_name:
            raise ValueError(
                "The two tables don't appear to have the same name."
            )

        #######################################################################

        # Because we're using sets here, the order is indeterminate. We sort
        # them, otherwise it's difficult to write good unit tests if the order
        # constantly changes.

        add_columns = [
            AddColumn(
                table_class_name=self.class_name,
                column_name=i.column._meta.name,
                db_column_name=i.column._meta.db_column_name,
                column_class_name=i.column.__class__.__name__,
                column_class=i.column.__class__,
                params=i.column._meta.params,
            )
            for i in sorted(
                {ColumnComparison(column=column) for column in self.columns}
                - {
                    ColumnComparison(column=column) for column in value.columns
                },
                key=lambda x: x.column._meta.name,
            )
        ]

        drop_columns = [
            DropColumn(
                table_class_name=self.class_name,
                column_name=i.column._meta.name,
                db_column_name=i.column._meta.db_column_name,
                tablename=value.tablename,
            )
            for i in sorted(
                {ColumnComparison(column=column) for column in value.columns}
                - {ColumnComparison(column=column) for column in self.columns},
                key=lambda x: x.column._meta.name,
            )
        ]

        #######################################################################

        alter_columns: t.List[AlterColumn] = []

        for existing_column in value.columns:
            column = self.columns_map.get(existing_column._meta.name)
            if not column:
                # This is a new column - already captured above.
                continue

            delta = compare_dicts(
                serialise_params(column._meta.params).params,
                serialise_params(existing_column._meta.params).params,
            )

            old_params = {
                key: existing_column._meta.params.get(key)
                for key, _ in delta.items()
            }

            if delta or (column.__class__ != existing_column.__class__):
                alter_columns.append(
                    AlterColumn(
                        table_class_name=self.class_name,
                        tablename=self.tablename,
                        column_name=column._meta.name,
                        db_column_name=column._meta.db_column_name,
                        params=deserialise_params(delta),
                        old_params=old_params,
                        column_class=column.__class__,
                        old_column_class=existing_column.__class__,
                    )
                )

        return TableDelta(
            add_columns=add_columns,
            drop_columns=drop_columns,
            alter_columns=alter_columns,
        )

    def __hash__(self) -> int:
        """
        We have to return an integer, which is why convert the string this way.
        """
        return hash(self.class_name + self.tablename)

    def __eq__(self, value) -> bool:
        """
        This is used by sets for uniqueness checks.
        """
        if not isinstance(value, DiffableTable):
            return False
        return (self.class_name == value.class_name) and (
            self.tablename == value.tablename
        )

    def __str__(self):
        return f"{self.class_name} - {self.tablename}"

    def to_table_class(self) -> t.Type[Table]:
        """
        Converts the DiffableTable into a Table subclass.
        """
        return create_table_class(
            class_name=self.class_name,
            class_kwargs={"tablename": self.tablename},
            class_members={
                column._meta.name: column for column in self.columns
            },
        )



import decimal
import uuid
import warnings
from enum import Enum
from unittest import TestCase

import pytest

from piccolo.apps.migrations.auto.serialisation import (
    CanConflictWithGlobalNames,
    Import,
    UniqueGlobalNameConflictWarning,
    UniqueGlobalNames,
    UniqueGlobalNamesMeta,
    serialise_params,
)
from piccolo.columns.base import OnDelete
from piccolo.columns.choices import Choice
from piccolo.columns.column_types import Varchar
from piccolo.columns.defaults import UUID4, DateNow, TimeNow, TimestampNow
from piccolo.columns.reference import LazyTableReference
from tests.base import engine_is


class TestUniqueGlobalNamesMeta:
    def test_duplicate_class_attribute_values_raises_error(self):
        with pytest.raises(ValueError):

            class IncorrectUniqueGlobalNames(metaclass=UniqueGlobalNamesMeta):
                A = "duplicate"
                B = "duplicate"


class TestUniqueGlobals:
    def test_contains_column_types(self):
        assert getattr(UniqueGlobalNames, "COLUMN_VARCHAR", "Varchar")
        assert getattr(UniqueGlobalNames, "COLUMN_SECRET", "Secret")
        assert getattr(UniqueGlobalNames, "COLUMN_TEXT", "Text")
        assert getattr(UniqueGlobalNames, "COLUMN_UUID", "UUID")
        assert getattr(UniqueGlobalNames, "COLUMN_INTEGER", "Integer")
        assert getattr(UniqueGlobalNames, "COLUMN_BIGINT", "BigInt")
        assert getattr(UniqueGlobalNames, "COLUMN_SMALLINT", "SmallInt")
        assert getattr(UniqueGlobalNames, "COLUMN_SERIAL", "Serial")
        assert getattr(UniqueGlobalNames, "COLUMN_BIGSERIAL", "BigSerial")
        assert getattr(UniqueGlobalNames, "COLUMN_PRIMARYKEY", "PrimaryKey")
        assert getattr(UniqueGlobalNames, "COLUMN_TIMESTAMP", "Timestamp")
        assert getattr(UniqueGlobalNames, "COLUMN_TIMESTAMPZ", "Timestampz")
        assert getattr(UniqueGlobalNames, "COLUMN_DATE", "Date")
        assert getattr(UniqueGlobalNames, "COLUMN_TIME", "Time")
        assert getattr(UniqueGlobalNames, "COLUMN_INTERVAL", "Interval")
        assert getattr(UniqueGlobalNames, "COLUMN_BOOLEAN", "Boolean")
        assert getattr(UniqueGlobalNames, "COLUMN_NUMERIC", "Numeric")
        assert getattr(UniqueGlobalNames, "COLUMN_DECIMAL", "Decimal")
        assert getattr(UniqueGlobalNames, "COLUMN_FLOAT", "Float")
        assert getattr(
            UniqueGlobalNames, "COLUMN_DOUBLEPERCISION", "DoublePrecision"
        )
        assert getattr(UniqueGlobalNames, "COLUMN_FOREIGNKEY", "ForeignKey")
        assert getattr(UniqueGlobalNames, "COLUMN_JSON", "JSON")
        assert getattr(UniqueGlobalNames, "COLUMN_BYTEA", "Bytea")
        assert getattr(UniqueGlobalNames, "COLUMN_BLOB", "Blob")
        assert getattr(UniqueGlobalNames, "COLUMN_ARRAY", "Array")

    def test_warn_if_is_conflicting_name(self):
        with pytest.warns(None) as recorded_warnings:
            UniqueGlobalNames.warn_if_is_conflicting_name(
                "SuperMassiveBlackHole"
            )

            if len(recorded_warnings) != 0:
                pytest.fail("Unexpected warning!")

        with pytest.warns(
            UniqueGlobalNameConflictWarning
        ) as recorded_warnings:
            UniqueGlobalNames.warn_if_is_conflicting_name("Varchar")

            if len(recorded_warnings) != 1:
                pytest.fail("Expected 1 warning!")

    def test_is_conflicting_name(self):
        assert (
            UniqueGlobalNames.is_conflicting_name("SuperMassiveBlackHole")
            is False
        )
        assert UniqueGlobalNames.is_conflicting_name("Varchar") is True

    def test_warn_if_are_conflicting_objects(self):
        class ConflictingCls1(CanConflictWithGlobalNames):
            def warn_if_is_conflicting_with_global_name(self):
                pass

        class ConflictingCls2(CanConflictWithGlobalNames):
            def warn_if_is_conflicting_with_global_name(self):
                pass

        class ConflictingCls3(CanConflictWithGlobalNames):
            def warn_if_is_conflicting_with_global_name(self):
                warnings.warn("test", UniqueGlobalNameConflictWarning)

        with pytest.warns(None) as recorded_warnings:
            UniqueGlobalNames.warn_if_are_conflicting_objects(
                [ConflictingCls1(), ConflictingCls2()]
            )

            if len(recorded_warnings) != 0:
                pytest.fail("Unexpected warning!")

        with pytest.warns(
            UniqueGlobalNameConflictWarning
        ) as recorded_warnings:
            UniqueGlobalNames.warn_if_are_conflicting_objects(
                [ConflictingCls2(), ConflictingCls3()]
            )

            if len(recorded_warnings) != 1:
                pytest.fail("Expected 1 warning!")


class TestImport:
    def test_with_module_only(self):
        assert repr(Import(module="a.b.c")) == "import a.b.c"

    def test_with_module_and_target(self):
        assert repr(Import(module="a.b", target="c")) == "from a.b import c"

    def test_warn_if_is_conflicting_with_global_name_with_module_only(self):
        with pytest.warns(None) as recorded_warnings:
            Import(module="a.b.c").warn_if_is_conflicting_with_global_name()

            if len(recorded_warnings) != 0:
                pytest.fail("Unexpected warning!")

        with pytest.warns(
            UniqueGlobalNameConflictWarning
        ) as recorded_warnings:
            Import(module="Varchar").warn_if_is_conflicting_with_global_name()

            if len(recorded_warnings) != 1:
                pytest.fail("Expected 1 warning!")

        with pytest.warns(None) as recorded_warnings:
            Import(
                module="Varchar", expect_conflict_with_global_name="Varchar"
            ).warn_if_is_conflicting_with_global_name()

            if len(recorded_warnings) != 0:
                pytest.fail("Unexpected warning!")

    def test_warn_if_is_conflicting_with_global_name_with_module_and_target(
        self,
    ):
        with pytest.warns(None) as recorded_warnings:
            Import(
                module="a.b", target="c"
            ).warn_if_is_conflicting_with_global_name()

            if len(recorded_warnings) != 0:
                pytest.fail("Unexpected warning!")

        with pytest.warns(
            UniqueGlobalNameConflictWarning
        ) as recorded_warnings:
            Import(
                module="a.b", target="Varchar"
            ).warn_if_is_conflicting_with_global_name()

            if len(recorded_warnings) != 1:
                pytest.fail("Expected 1 warning!")

        with pytest.warns(None) as recorded_warnings:
            Import(
                module="a.b",
                target="Varchar",
                expect_conflict_with_global_name="Varchar",
            ).warn_if_is_conflicting_with_global_name()

            if len(recorded_warnings) != 0:
                pytest.fail("Unexpected warning!")


def example_function():
    pass


class TestSerialiseParams(TestCase):
    def test_time(self):
        serialised = serialise_params(params={"default": TimeNow()})
        self.assertEqual(serialised.params["default"].__repr__(), "TimeNow()")
        self.assertTrue(len(serialised.extra_imports) == 1)
        self.assertEqual(
            serialised.extra_imports[0].__str__(),
            "from piccolo.columns.defaults.time import TimeNow",
        )

    def test_date(self):
        serialised = serialise_params(params={"default": DateNow()})
        self.assertEqual(serialised.params["default"].__repr__(), "DateNow()")

    def test_timestamp(self):
        serialised = serialise_params(params={"default": TimestampNow()})
        self.assertTrue(
            serialised.params["default"].__repr__() == "TimestampNow()"
        )

    def test_uuid(self):
        serialised = serialise_params(params={"default": uuid.UUID(int=4)})
        assert (
            repr(serialised.params["default"])
            == 'uuid.UUID("00000000-0000-0000-0000-000000000004")'
        )

        serialised = serialise_params(params={"default": UUID4()})
        self.assertTrue(serialised.params["default"].__repr__() == "UUID4()")

    def test_decimal(self):
        serialised = serialise_params(
            params={"default": decimal.Decimal("1.2")}
        )
        assert repr(serialised.params["default"]) == 'decimal.Decimal("1.2")'

    def test_lazy_table_reference(self):
        # These are equivalent:
        references_list = [
            LazyTableReference(table_class_name="Manager", app_name="music"),
            LazyTableReference(
                table_class_name="Manager",
                module_path="tests.example_apps.music.tables",
            ),
        ]

        for references in references_list:
            serialised = serialise_params(params={"references": references})
            self.assertTrue(
                serialised.params["references"].__repr__() == "Manager"
            )

            self.assertTrue(len(serialised.extra_imports) == 1)
            self.assertEqual(
                serialised.extra_imports[0].__str__(),
                "from piccolo.table import Table",
            )

            self.assertTrue(len(serialised.extra_definitions) == 1)

            if engine_is("postgres"):
                self.assertEqual(
                    serialised.extra_definitions[0].__str__(),
                    (
                        'class Manager(Table, tablename="manager"): '
                        "id = Serial(null=False, primary_key=True, unique=False, "  # noqa: E501
                        "index=False, index_method=IndexMethod.btree, "
                        "choices=None, db_column_name='id', secret=False)"
                    ),
                )

            if engine_is("cockroach"):
                self.assertEqual(
                    serialised.extra_definitions[0].__str__(),
                    (
                        'class Manager(Table, tablename="manager"): '
                        "id = Serial(null=False, primary_key=True, unique=False, "  # noqa: E501
                        "index=False, index_method=IndexMethod.btree, "
                        "choices=None, db_column_name='id', secret=False)"
                    ),
                )

    def test_function(self):
        serialised = serialise_params(params={"default": example_function})
        self.assertTrue(
            serialised.params["default"].__repr__() == "example_function"
        )

        self.assertTrue(len(serialised.extra_imports) == 1)
        self.assertEqual(
            serialised.extra_imports[0].__str__(),
            (
                "from tests.apps.migrations.auto.test_serialisation import "
                "example_function"
            ),
        )

        self.assertTrue(len(serialised.extra_definitions) == 0)

    def test_lambda(self):
        """
        Make sure lambda functions are rejected.
        """
        with self.assertRaises(ValueError) as manager:
            serialise_params(params={"default": lambda x: x + 1})

        self.assertEqual(
            manager.exception.__str__(), "Lambdas can't be serialised"
        )

    def test_builtins(self):
        """
        Make sure builtins can be serialised properly.
        """
        serialised = serialise_params(params={"default": list})
        self.assertTrue(serialised.params["default"].__repr__() == "list")

        self.assertTrue(len(serialised.extra_imports) == 0)

    def test_column_instance(self):
        """
        Make sure Column instances can be serialised properly. An example
        use case is when a `base_column` argument is passed to an `Array`
        column.
        """
        serialised = serialise_params(params={"base_column": Varchar()})

        self.assertEqual(
            serialised.params["base_column"].__repr__(),
            "Varchar(length=255, default='', null=False, primary_key=False, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name=None, secret=False)",  # noqa: E501
        )

        self.assertEqual(
            {i.__repr__() for i in serialised.extra_imports},
            {
                "from piccolo.columns.column_types import Varchar",
                "from piccolo.columns.indexes import IndexMethod",
            },
        )

    def test_enum_type(self):
        """
        Make sure Enum types can be serialised properly.
        """

        class Choices(Enum):
            a = 1
            b = 2
            c = Choice(value=3, display_name="c1")

        serialised = serialise_params(params={"choices": Choices})

        self.assertEqual(
            serialised.params["choices"].__repr__(),
            "Enum('Choices', {'a': 1, 'b': 2, 'c': Choice(value=3, display_name='c1')})",  # noqa: E501
        )

        self.assertEqual(
            {i.__repr__() for i in serialised.extra_imports},
            {
                "from piccolo.columns.choices import Choice",
                "from enum import Enum",
            },
        )

    def test_custom_enum_instance(self):
        """
        Make sure custom Enum instances can be serialised properly. An example
        is when a user defines a choices Enum, and then sets the default to
        one of those choices.
        """

        class Choices(Enum):
            a = 1
            b = 2

        serialised = serialise_params(params={"default": Choices.a})

        self.assertEqual(serialised.params["default"], 1)
        self.assertEqual(serialised.extra_imports, [])
        self.assertEqual(serialised.extra_definitions, [])

    def test_builtin_enum_instance(self):
        """
        Make sure Enum instances defiend in Piccolo can be serialised properly
        - for example, with on_delete.
        """
        serialised = serialise_params(params={"on_delete": OnDelete.cascade})

        self.assertEqual(
            serialised.params["on_delete"].__repr__(), "OnDelete.cascade"
        )
        self.assertEqual(
            [i.__repr__() for i in serialised.extra_imports],
            ["from piccolo.columns.base import OnDelete"],
        )
        self.assertEqual(serialised.extra_definitions, [])



"""
Credit to the Django Extensions team for inspiring this tool.
"""

import dataclasses
import os
import sys
import typing as t

import jinja2

from piccolo.conf.apps import Finder

TEMPLATE_DIRECTORY = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "templates"
)

JINJA_ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader(searchpath=TEMPLATE_DIRECTORY),
)


@dataclasses.dataclass
class GraphColumn:
    name: str
    type: str


@dataclasses.dataclass
class GraphTable:
    name: str
    columns: t.List[GraphColumn]


@dataclasses.dataclass
class GraphRelation:
    table_a: str
    table_b: str
    label: str


def render_template(**kwargs):
    template = JINJA_ENV.get_template("graphviz.dot.jinja")
    return template.render(**kwargs)


def graph(
    apps: str = "all", direction: str = "LR", output: t.Optional[str] = None
):
    """
    Prints out a graphviz .dot file for your schema.

    :param apps:
        The name of the apps to include. If 'all' is given then every app is
        included. To specify multiple app names, separate them with commas.
        For example --apps="app1,app2".
    :param direction:
        How the tables should be orientated - by default it's "LR" which is
        left to right, so the graph will be landscape. The alternative is
        "TB", which is top to bottom, so the graph will be portrait.
    :param output:
        If specified, rather than printing out the file contents, they'll be
        written to this file. For example --output=graph.dot

    """
    finder = Finder()
    app_names = finder.get_sorted_app_names()

    if apps != "all":
        given_app_names = [i.strip() for i in apps.split(",")]
        delta = set(given_app_names) - set(app_names)
        if delta:
            sys.exit(f"These apps aren't recognised: {', '.join(delta)}.")
        app_names = given_app_names

    tables: t.List[GraphTable] = []
    relations: t.List[GraphRelation] = []

    for app_name in app_names:
        app_config = finder.get_app_config(app_name=app_name)
        for table_class in app_config.table_classes:
            tables.append(
                GraphTable(
                    name=table_class.__name__,
                    columns=[
                        GraphColumn(
                            name=i._meta.name, type=i.__class__.__name__
                        )
                        for i in table_class._meta.columns
                    ],
                )
            )
            for fk_column in table_class._meta.foreign_key_columns:
                reference_table_class = (
                    fk_column._foreign_key_meta.resolved_references
                )
                relations.append(
                    GraphRelation(
                        table_a=table_class.__name__,
                        table_b=reference_table_class.__name__,
                        label=fk_column._meta.name,
                    )
                )

    contents = render_template(
        tables=tables, relations=relations, direction=direction
    )

    if output is None:
        print(contents)
    else:
        with open(output, "w") as f:
            f.write(contents)



import datetime
import decimal
import uuid
from unittest import TestCase

from piccolo.columns.column_types import (
    UUID,
    UUID4,
    BigInt,
    Date,
    DateNow,
    DoublePrecision,
    ForeignKey,
    Integer,
    Numeric,
    Real,
    SmallInt,
    Text,
    Time,
    TimeNow,
    Timestamp,
    TimestampNow,
    Varchar,
)
from piccolo.table import Table


class TestDefaults(TestCase):
    """
    Columns check the type of the default argument.
    """

    def test_int(self):
        for _type in (Integer, BigInt, SmallInt):
            _type(default=0)
            _type(default=None, null=True)
            with self.assertRaises(ValueError):
                _type(default="hello world")

    def test_text(self):
        for _type in (Text, Varchar):
            _type(default="")
            _type(default=None, null=True)
            with self.assertRaises(ValueError):
                _type(default=123)

    def test_real(self):
        Real(default=0.0)
        Real(default=None, null=True)
        with self.assertRaises(ValueError):
            Real(default="hello world")

    def test_double_precision(self):
        DoublePrecision(default=0.0)
        DoublePrecision(default=None, null=True)
        with self.assertRaises(ValueError):
            DoublePrecision(default="hello world")

    def test_numeric(self):
        Numeric(default=decimal.Decimal(1.0))
        Numeric(default=None, null=True)
        with self.assertRaises(ValueError):
            Numeric(default="hello world")

    def test_uuid(self):
        UUID(default=None, null=True)
        UUID(default=UUID4())
        UUID(default=uuid.uuid4())
        with self.assertRaises(ValueError):
            UUID(default="hello world")

    def test_time(self):
        Time(default=None, null=True)
        Time(default=TimeNow())
        Time(default=datetime.datetime.now().time())
        with self.assertRaises(ValueError):
            Time(default="hello world")

    def test_date(self):
        Date(default=None, null=True)
        Date(default=DateNow())
        Date(default=datetime.datetime.now().date())
        with self.assertRaises(ValueError):
            Date(default="hello world")

    def test_timestamp(self):
        Timestamp(default=None, null=True)
        Timestamp(default=TimestampNow())
        Timestamp(default=datetime.datetime.now())
        with self.assertRaises(ValueError):
            Timestamp(default="hello world")

    def test_foreignkey(self):
        class MyTable(Table):
            pass

        ForeignKey(references=MyTable, default=None, null=True)
        ForeignKey(references=MyTable, default=1)
        with self.assertRaises(ValueError):
            ForeignKey(references=MyTable, default="hello world")



from __future__ import annotations

import datetime
import typing as t
from enum import Enum

from .base import Default


class TimestampOffset(Default):
    def __init__(
        self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0
    ):
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

    @property
    def postgres(self):
        interval_string = self.get_postgres_interval_string(
            ["days", "hours", "minutes", "seconds"]
        )
        return f"CURRENT_TIMESTAMP + INTERVAL '{interval_string}'"

    @property
    def cockroach(self):
        interval_string = self.get_postgres_interval_string(
            ["days", "hours", "minutes", "seconds"]
        )
        return f"CURRENT_TIMESTAMP::TIMESTAMP + INTERVAL '{interval_string}'"

    @property
    def sqlite(self):
        interval_string = self.get_sqlite_interval_string(
            ["days", "hours", "minutes", "seconds"]
        )
        return f"(datetime(CURRENT_TIMESTAMP, {interval_string}))"

    def python(self):
        return datetime.datetime.now() + datetime.timedelta(
            days=self.days,
            hours=self.hours,
            minutes=self.minutes,
            seconds=self.seconds,
        )


class TimestampNow(Default):
    @property
    def postgres(self):
        return "current_timestamp"

    @property
    def cockroach(self):
        return "current_timestamp::TIMESTAMP"

    @property
    def sqlite(self):
        return "current_timestamp"

    def python(self):
        return datetime.datetime.now()


class TimestampCustom(Default):
    def __init__(
        self,
        year: int = 2000,
        month: int = 1,
        day: int = 1,
        hour: int = 0,
        second: int = 0,
        microsecond: int = 0,
    ):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.second = second
        self.microsecond = microsecond

    @property
    def datetime(self):
        return datetime.datetime(
            year=self.year,
            month=self.month,
            day=self.day,
            hour=self.hour,
            second=self.second,
            microsecond=self.microsecond,
        )

    @property
    def postgres(self):
        return "'{}'".format(self.datetime.isoformat().replace("T", " "))

    @property
    def cockroach(self):
        return "'{}'::TIMESTAMP".format(
            self.datetime.isoformat().replace("T", " ")
        )

    @property
    def sqlite(self):
        return "'{}'".format(self.datetime.isoformat().replace("T", " "))

    def python(self):
        return self.datetime

    @classmethod
    def from_datetime(cls, instance: datetime.datetime):  # type: ignore
        return cls(
            year=instance.year,
            month=instance.month,
            day=instance.month,
            hour=instance.hour,
            second=instance.second,
            microsecond=instance.microsecond,
        )


###############################################################################
# For backwards compatibility:


class DatetimeDefault:
    now = TimestampNow()


###############################################################################

TimestampArg = t.Union[
    TimestampCustom,
    TimestampNow,
    TimestampOffset,
    Enum,
    None,
    datetime.datetime,
    DatetimeDefault,
]


__all__ = [
    "TimestampArg",
    "TimestampCustom",
    "TimestampNow",
    "TimestampOffset",
]



from __future__ import annotations

import copy
import datetime
import decimal
import inspect
import typing as t
import uuid
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field, fields
from enum import Enum

from piccolo.columns.choices import Choice
from piccolo.columns.combination import Where
from piccolo.columns.defaults.base import Default
from piccolo.columns.defaults.interval import IntervalCustom
from piccolo.columns.indexes import IndexMethod
from piccolo.columns.operators.comparison import (
    ComparisonOperator,
    Equal,
    GreaterEqualThan,
    GreaterThan,
    ILike,
    In,
    IsNotNull,
    IsNull,
    LessEqualThan,
    LessThan,
    Like,
    NotEqual,
    NotIn,
    NotLike,
)
from piccolo.columns.reference import LazyTableReference
from piccolo.utils.warnings import colored_warning

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns.column_types import ForeignKey
    from piccolo.table import Table


class OnDelete(str, Enum):
    """
    Used by :class:`ForeignKey <piccolo.columns.column_types.ForeignKey>` to
    specify the behaviour when a related row is deleted.
    """

    cascade = "CASCADE"
    restrict = "RESTRICT"
    no_action = "NO ACTION"
    set_null = "SET NULL"
    set_default = "SET DEFAULT"

    def __str__(self):
        return f"{self.__class__.__name__}.{self.name}"

    def __repr__(self):
        return self.__str__()


class OnUpdate(str, Enum):
    """
    Used by :class:`ForeignKey <piccolo.columns.column_types.ForeignKey>` to
    specify the behaviour when a related row is updated.
    """

    cascade = "CASCADE"
    restrict = "RESTRICT"
    no_action = "NO ACTION"
    set_null = "SET NULL"
    set_default = "SET DEFAULT"

    def __str__(self):
        return f"{self.__class__.__name__}.{self.name}"

    def __repr__(self):
        return self.__str__()


@dataclass
class ForeignKeyMeta:
    references: t.Union[t.Type[Table], LazyTableReference]
    on_delete: OnDelete
    on_update: OnUpdate
    target_column: t.Union[Column, str, None]
    proxy_columns: t.List[Column] = field(default_factory=list)

    @property
    def resolved_references(self) -> t.Type[Table]:
        """
        Evaluates the ``references`` attribute if it's a ``LazyTableReference``,
        raising a ``ValueError`` if it fails, otherwise returns a ``Table``
        subclass.
        """  # noqa: E501
        from piccolo.table import Table

        if isinstance(self.references, LazyTableReference):
            return self.references.resolve()
        elif inspect.isclass(self.references) and issubclass(
            self.references, Table
        ):
            return self.references
        else:
            raise ValueError(
                "The references attribute is neither a Table subclass or a "
                "LazyTableReference instance."
            )

    @property
    def resolved_target_column(self) -> Column:
        if self.target_column is None:
            return self.resolved_references._meta.primary_key
        elif isinstance(self.target_column, Column):
            return self.resolved_references._meta.get_column_by_name(
                self.target_column._meta.name
            )
        elif isinstance(self.target_column, str):
            return self.resolved_references._meta.get_column_by_name(
                self.target_column
            )
        else:
            raise ValueError("Unable to resolve target_column.")

    def copy(self) -> ForeignKeyMeta:
        kwargs = self.__dict__.copy()
        kwargs.update(proxy_columns=self.proxy_columns.copy())
        return self.__class__(**kwargs)

    def __copy__(self) -> ForeignKeyMeta:
        return self.copy()

    def __deepcopy__(self, memo) -> ForeignKeyMeta:
        """
        We override deepcopy, as it's too slow if it has to recreate
        everything.
        """
        return self.copy()


@dataclass
class ColumnMeta:
    """
    We store as many attributes in ColumnMeta as possible, to help avoid name
    clashes with user defined attributes.
    """

    # General attributes:
    null: bool = False
    primary_key: bool = False
    unique: bool = False
    index: bool = False
    index_method: IndexMethod = IndexMethod.btree
    required: bool = False
    help_text: t.Optional[str] = None
    choices: t.Optional[t.Type[Enum]] = None
    secret: bool = False
    auto_update: t.Any = ...

    # Used for representing the table in migrations and the playground.
    params: t.Dict[str, t.Any] = field(default_factory=dict)

    ###########################################################################

    # Lets you to map a column to a database column with a different name.
    _db_column_name: t.Optional[str] = None

    @property
    def db_column_name(self) -> str:
        return self._db_column_name or self.name

    @db_column_name.setter
    def db_column_name(self, value: str):
        self._db_column_name = value

    ###########################################################################

    # Set by the Table Metaclass:
    _name: t.Optional[str] = None
    _table: t.Optional[t.Type[Table]] = None

    @property
    def name(self) -> str:
        if not self._name:
            raise ValueError(
                "`_name` isn't defined - the Table Metaclass should set it."
            )
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def table(self) -> t.Type[Table]:
        if not self._table:
            raise ValueError(
                "`_table` isn't defined - the Table Metaclass should set it."
            )
        return self._table

    ###########################################################################

    # Used by Foreign Keys:
    call_chain: t.List["ForeignKey"] = field(default_factory=list)
    table_alias: t.Optional[str] = None

    ###########################################################################

    @property
    def engine_type(self) -> str:
        engine = self.table._meta.db
        if engine:
            return engine.engine_type
        else:
            raise ValueError("The table has no engine defined.")

    def get_choices_dict(self) -> t.Optional[t.Dict[str, t.Any]]:
        """
        Return the choices Enum as a dict. It maps the attribute name to a
        dict containing the display name, and value.
        """
        if self.choices is None:
            return None
        output = {}
        for element in self.choices:
            if isinstance(element.value, Choice):
                display_name = element.value.display_name
                value = element.value.value
            else:
                display_name = element.name.replace("_", " ").title()
                value = element.value

            output[element.name] = {
                "display_name": display_name,
                "value": value,
            }

        return output

    ###########################################################################

    def get_default_alias(self):
        column_name = self.db_column_name

        if self.call_chain:
            column_name = (
                "$".join(
                    t.cast(str, i._meta.db_column_name)
                    for i in self.call_chain
                )
                + f"${column_name}"
            )

        return column_name

    def _get_path(self, include_quotes: bool = False):
        column_name = self.db_column_name

        if self.call_chain:
            table_alias = self.call_chain[-1]._meta.table_alias
            if include_quotes:
                return f'"{table_alias}"."{column_name}"'
            else:
                return f"{table_alias}.{column_name}"
        else:
            if include_quotes:
                return f'"{self.table._meta.tablename}"."{column_name}"'
            else:
                return f"{self.table._meta.tablename}.{column_name}"

    def get_full_name(
        self, with_alias: bool = True, include_quotes: bool = True
    ) -> str:
        """
        Returns the full column name, taking into account joins.

        :param with_alias:
            Examples:

            .. code-block python::

                >>> Band.manager.name._meta.get_full_name(with_alias=False)
                'band$manager.name'

                >>> Band.manager.name._meta.get_full_name(with_alias=True)
                'band$manager.name AS "manager$name"'

        :param include_quotes:
            If you're using the name in a SQL query, each component needs to be
            surrounded by double quotes, in case the table or column name
            clashes with a reserved SQL keyword (for example, a column called
            ``order``).

            .. code-block python::

                >>> column._meta.get_full_name(include_quotes=True)
                '"my_table_name"."my_column_name"'

                >>> column._meta.get_full_name(include_quotes=False)
                'my_table_name.my_column_name'


        """
        full_name = self._get_path(include_quotes=include_quotes)

        if with_alias and self.call_chain:
            alias = self.get_default_alias()
            if include_quotes:
                full_name += f' AS "{alias}"'
            else:
                full_name += f" AS {alias}"

        return full_name

    ###########################################################################

    def copy(self) -> ColumnMeta:
        kwargs = self.__dict__.copy()
        kwargs.update(
            params=self.params.copy(),
            call_chain=self.call_chain.copy(),
        )

        # Make sure we don't accidentally include any other attributes which
        # aren't supported by the constructor.
        field_names = [i.name for i in fields(self.__class__)]
        kwargs = {
            kwarg: value
            for kwarg, value in kwargs.items()
            if kwarg in field_names
        }

        return self.__class__(**kwargs)

    def __copy__(self) -> ColumnMeta:
        return self.copy()

    def __deepcopy__(self, memo) -> ColumnMeta:
        """
        We override deepcopy, as it's too slow if it has to recreate
        everything.
        """
        return self.copy()


class Selectable(metaclass=ABCMeta):
    """
    Anything which inherits from this can be used in a select query.
    """

    _alias: t.Optional[str]

    @abstractmethod
    def get_select_string(
        self, engine_type: str, with_alias: bool = True
    ) -> str:
        """
        In a query, what to output after the select statement - could be a
        column name, a sub query, a function etc. For a column it will be the
        column name.
        """
        raise NotImplementedError()

    def as_alias(self, alias: str) -> Selectable:
        """
        Allows column names to be changed in the result of a select.
        """
        self._alias = alias
        return self


class Column(Selectable):
    """
    All other columns inherit from ``Column``. Don't use it directly.

    The following arguments apply to all column types:

    :param null:
        Whether the column is nullable.

    :param primary_key:
        If set, the column is used as a primary key.

    :param default:
        The column value to use if not specified by the user.

    :param unique:
        If set, a unique constraint will be added to the column.

    :param index:
        Whether an index is created for the column, which can improve
        the speed of selects, but can slow down inserts.

    :param index_method:
        If index is set to ``True``, this specifies what type of index is
        created.

    :param required:
        This isn't used by the database - it's to indicate to other tools that
        the user must provide this value. Example uses are in serialisers for
        API endpoints, and form fields.

    :param help_text:
        This provides some context about what the column is being used for. For
        example, for a ``Decimal`` column called ``value``, it could say
        ``'The units are millions of dollars'``. The database doesn't use this
        value, but tools such as Piccolo Admin use it to show a tooltip in the
        GUI.

    :param choices:
        An optional Enum - when specified, other tools such as Piccolo Admin
        will render the available options in the GUI.

    :param db_column_name:
        If specified, you can override the name used for the column in the
        database. The main reason for this is when using a legacy database,
        with a problematic column name (for example ``'class'``, which is a
        reserved Python keyword). Here's an example:

        .. code-block:: python

            class MyTable(Table):
                class_ = Varchar(db_column_name="class")

            >>> await MyTable.select(MyTable.class_)
            [{'id': 1, 'class': 'test'}]

        This is an advanced feature which you should only need in niche
        situations.

    :param secret:
        If ``secret=True`` is specified, it allows a user to automatically
        omit any fields when doing a select query, to help prevent
        inadvertent leakage of sensitive data.

        .. code-block:: python

            class Band(Table):
                name = Varchar()
                net_worth = Integer(secret=True)

            >>> await Band.select(exclude_secrets=True)
            [{'name': 'Pythonistas'}]

    :param auto_update:
        Allows you to specify a value to set this column to each time it is
        updated (via ``MyTable.update``, or ``MyTable.save`` on an existing
        row). A common use case is having a ``modified_on`` column.

        .. code-block:: python

            class Band(Table):
                name = Varchar()
                popularity = Integer()
                # The value can be a function or static value:
                modified_on = Timestamp(auto_update=datetime.datetime.now)

            # This will automatically set the `modified_on` column to the
            # current timestamp, without having to explicitly set it:
            >>> await Band.update({
            ...     Band.popularity: Band.popularity + 100
            ... }).where(Band.name == 'Pythonistas')

        Note - this feature is implemented purely within the ORM. If you want
        similar functionality on the database level (i.e. if you plan on using
        raw SQL to perform updates), then you may be better off creating SQL
        triggers instead.

    """

    value_type: t.Type = int

    def __init__(
        self,
        null: bool = False,
        primary_key: bool = False,
        unique: bool = False,
        index: bool = False,
        index_method: IndexMethod = IndexMethod.btree,
        required: bool = False,
        help_text: t.Optional[str] = None,
        choices: t.Optional[t.Type[Enum]] = None,
        db_column_name: t.Optional[str] = None,
        secret: bool = False,
        auto_update: t.Any = ...,
        **kwargs,
    ) -> None:
        # This is for backwards compatibility - originally there were two
        # separate arguments `primary` and `key`, but they have now been merged
        # into `primary_key`.
        if (kwargs.get("primary") is True) and (kwargs.get("key") is True):
            primary_key = True

        # Used for migrations.
        # We deliberately omit 'required', 'auto_update' and 'help_text' as
        # they don't effect the actual schema.
        # 'choices' isn't used directly in the schema, but may be important
        # for data migrations.
        kwargs.update(
            {
                "null": null,
                "primary_key": primary_key,
                "unique": unique,
                "index": index,
                "index_method": index_method,
                "choices": choices,
                "db_column_name": db_column_name,
                "secret": secret,
            }
        )

        if choices is not None:
            self._validate_choices(choices, allowed_type=self.value_type)

        self._meta = ColumnMeta(
            null=null,
            primary_key=primary_key,
            unique=unique,
            index=index,
            index_method=index_method,
            params=kwargs,
            required=required,
            help_text=help_text,
            choices=choices,
            _db_column_name=db_column_name,
            secret=secret,
            auto_update=auto_update,
        )

        self._alias: t.Optional[str] = None

    def _validate_default(
        self,
        default: t.Any,
        allowed_types: t.Iterable[t.Union[None, t.Type[t.Any]]],
        allow_recursion: bool = True,
    ) -> bool:
        """
        Make sure that the default value is of the allowed types.
        """
        if getattr(self, "_validated", None):
            # If it has previously been validated by a subclass, don't
            # validate again.
            return True
        elif (
            default is None
            and None in allowed_types
            or type(default) in allowed_types
        ):
            self._validated = True
            return True
        elif callable(default):
            # We need to prevent recursion, otherwise a function which returns
            # a function would be an infinite loop.
            if allow_recursion and self._validate_default(
                default(), allowed_types=allowed_types, allow_recursion=False
            ):
                self._validated = True
                return True
        elif (
            isinstance(default, Enum) and type(default.value) in allowed_types
        ):
            self._validated = True
            return True

        raise ValueError(
            f"The default {default} isn't one of the permitted types - "
            f"{allowed_types}"
        )

    def _validate_choices(
        self, choices: t.Type[Enum], allowed_type: t.Type[t.Any]
    ) -> bool:
        """
        Make sure the choices value has values of the allowed_type.
        """
        if getattr(self, "_validated_choices", None):
            # If it has previously been validated by a subclass, don't
            # validate again.
            return True

        for element in choices:
            if isinstance(element.value, allowed_type):
                continue
            elif isinstance(element.value, Choice) and isinstance(
                element.value.value, allowed_type
            ):
                continue
            else:
                raise ValueError(
                    f"{element.name} doesn't have the correct type"
                )

        self._validated_choices = True

        return True

    def is_in(self, values: t.List[t.Any]) -> Where:
        if len(values) == 0:
            raise ValueError(
                "The `values` list argument must contain at least one value."
            )
        return Where(column=self, values=values, operator=In)

    def not_in(self, values: t.List[t.Any]) -> Where:
        if len(values) == 0:
            raise ValueError(
                "The `values` list argument must contain at least one value."
            )
        return Where(column=self, values=values, operator=NotIn)

    def like(self, value: str) -> Where:
        """
        Both SQLite and Postgres support LIKE, but they mean different things.

        In Postgres, LIKE is case sensitive (i.e. 'foo' equals 'foo', but
        'foo' doesn't equal 'Foo').

        In SQLite, LIKE is case insensitive for ASCII characters
        (i.e. 'foo' equals 'Foo'). But not for non-ASCII characters. To learn
        more, see the docs:

        https://sqlite.org/lang_expr.html#the_like_glob_regexp_and_match_operators

        """
        return Where(column=self, value=value, operator=Like)

    def ilike(self, value: str) -> Where:
        """
        Only Postgres supports ILIKE. It's used for case insensitive matching.

        For SQLite, it's just proxied to a LIKE query instead.

        """
        if self._meta.engine_type in ("postgres", "cockroach"):
            operator: t.Type[ComparisonOperator] = ILike
        else:
            colored_warning(
                "SQLite doesn't support ILIKE, falling back to LIKE."
            )
            operator = Like
        return Where(column=self, value=value, operator=operator)

    def not_like(self, value: str) -> Where:
        return Where(column=self, value=value, operator=NotLike)

    def __lt__(self, value) -> Where:
        return Where(column=self, value=value, operator=LessThan)

    def __le__(self, value) -> Where:
        return Where(column=self, value=value, operator=LessEqualThan)

    def __gt__(self, value) -> Where:
        return Where(column=self, value=value, operator=GreaterThan)

    def __ge__(self, value) -> Where:
        return Where(column=self, value=value, operator=GreaterEqualThan)

    def _equals(self, column: Column, including_joins: bool = False) -> bool:
        """
        We override ``__eq__``, in order to do queries such as:

        .. code-block:: python

            await Band.select().where(Band.name == 'Pythonistas')

        But this means that comparisons such as this can give unexpected
        results:

        .. code-block:: python

            # We would expect the answer to be `True`, but we get `Where`
            # instead:
            >>> MyTable.some_column == MyTable.some_column
            <Where>

        Also, column comparison is sometimes more complex than it appears. This
        is why we have this custom method for comparing columns.

        Take this example:

        .. code-block:: python

            Band.manager.name == Manager.name

        They both refer to the ``name`` column on the ``Manager`` table, except
        one has joins and the other doesn't.

        :param including_joins:
            If ``True``, then we check if the columns are the same, as well as
            their joins, i.e. ``Band.manager.name`` != ``Manager.name``.

        """
        if isinstance(column, Column):
            if (
                self._meta.name == column._meta.name
                and self._meta.table._meta.tablename
                == column._meta.table._meta.tablename
            ):
                if including_joins:
                    if len(column._meta.call_chain) == len(
                        self._meta.call_chain
                    ):
                        return all(
                            column_a._equals(column_b, including_joins=False)
                            for column_a, column_b in zip(
                                column._meta.call_chain,
                                self._meta.call_chain,
                            )
                        )

                else:
                    return True

        return False

    def __eq__(self, value) -> Where:  # type: ignore[override]
        if value is None:
            return Where(column=self, operator=IsNull)
        else:
            return Where(column=self, value=value, operator=Equal)

    def __ne__(self, value) -> Where:  # type: ignore[override]
        if value is None:
            return Where(column=self, operator=IsNotNull)
        else:
            return Where(column=self, value=value, operator=NotEqual)

    def __hash__(self):
        return hash(self._meta.name)

    def is_null(self) -> Where:
        """
        Can be used instead of ``MyTable.column == None``, because some linters
        don't like a comparison to ``None``.
        """
        return Where(column=self, operator=IsNull)

    def is_not_null(self) -> Where:
        """
        Can be used instead of ``MyTable.column != None``, because some linters
        don't like a comparison to ``None``.
        """
        return Where(column=self, operator=IsNotNull)

    def as_alias(self, name: str) -> Column:
        """
        Allows column names to be changed in the result of a select.

        For example:

        .. code-block:: python

            >>> await Band.select(Band.name.as_alias('title')).run()
            {'title': 'Pythonistas'}

        """
        column = copy.deepcopy(self)
        column._alias = name
        return column

    def join_on(self, column: Column) -> ForeignKey:
        """
        Joins are typically performed via foreign key columns. For example,
        here we get the band's name and the manager's name::

            class Manager(Table):
                name = Varchar()

            class Band(Table):
                name = Varchar()
                manager = ForeignKey(Manager)

            >>> await Band.select(Band.name, Band.manager.name)

        The ``join_on`` method lets you join tables even when foreign keys
        don't exist, by joining on a column in another table.

        For example, here we want to get the manager's email, but no foreign
        key exists::

            class Manager(Table):
                name = Varchar(unique=True)
                email = Varchar()

            class Band(Table):
                name = Varchar()
                manager_name = Varchar()

            >>> await Band.select(
            ...     Band.name,
            ...     Band.manager_name.join_on(Manager.name).email
            ... )

        """
        from piccolo.columns.column_types import ForeignKey

        virtual_foreign_key = ForeignKey(
            references=column._meta.table, target_column=column
        )
        virtual_foreign_key._meta._name = self._meta.name
        virtual_foreign_key._meta.call_chain = [*self._meta.call_chain]
        virtual_foreign_key._meta._table = self._meta.table
        virtual_foreign_key.set_proxy_columns()
        return virtual_foreign_key

    def get_default_value(self) -> t.Any:
        """
        If the column has a default attribute, return it. If it's callable,
        return the response instead.
        """
        default = getattr(self, "default", ...)
        if default is not ...:
            default = default.value if isinstance(default, Enum) else default
            is_callable = hasattr(default, "__call__")
            return default() if is_callable else default  # type: ignore
        return None

    def get_select_string(
        self, engine_type: str, with_alias: bool = True
    ) -> str:
        """
        How to refer to this column in a SQL query, taking account of any joins
        and aliases.
        """
        if with_alias:
            if self._alias:
                original_name = self._meta.get_full_name(
                    with_alias=False,
                )
                return f'{original_name} AS "{self._alias}"'
            else:
                return self._meta.get_full_name(
                    with_alias=True,
                )

        return self._meta.get_full_name(with_alias=False)

    def get_where_string(self, engine_type: str) -> str:
        return self.get_select_string(
            engine_type=engine_type, with_alias=False
        )

    def get_sql_value(self, value: t.Any) -> t.Any:
        """
        When using DDL statements, we can't parameterise the values. An example
        is when setting the default for a column. So we have to convert from
        the Python type to a string representation which we can include in our
        DDL statements.

        :param value:
            The Python value to convert to a string usable in a DDL statement
            e.g. 1.
        :returns:
            The string usable in the DDL statement e.g. '1'.

        """
        if isinstance(value, Default):
            return getattr(value, self._meta.engine_type)
        elif value is None:
            return "null"
        elif isinstance(value, (float, decimal.Decimal)):
            return str(value)
        elif isinstance(value, str):
            return f"'{value}'"
        elif isinstance(value, bool):
            return str(value).lower()
        elif isinstance(value, datetime.datetime):
            return f"'{value.isoformat().replace('T', ' ')}'"
        elif isinstance(value, datetime.date):
            return f"'{value.isoformat()}'"
        elif isinstance(value, datetime.time):
            return f"'{value.isoformat()}'"
        elif isinstance(value, datetime.timedelta):
            interval = IntervalCustom.from_timedelta(value)
            return getattr(interval, self._meta.engine_type)
        elif isinstance(value, bytes):
            return f"'{value.hex()}'"
        elif isinstance(value, uuid.UUID):
            return f"'{value}'"
        elif isinstance(value, list):
            # Convert to the array syntax.
            return (
                "'{"
                + ", ".join(
                    f'"{i}"'
                    if isinstance(i, str)
                    else str(self.get_sql_value(i))
                    for i in value
                )
            ) + "}'"
        else:
            return value

    @property
    def column_type(self):
        return self.__class__.__name__.upper()

    @property
    def ddl(self) -> str:
        """
        Used when creating tables.
        """
        query = f'"{self._meta.db_column_name}" {self.column_type}'
        if self._meta.primary_key:
            query += " PRIMARY KEY"
        if self._meta.unique:
            query += " UNIQUE"
        if not self._meta.null:
            query += " NOT NULL"

        foreign_key_meta: t.Optional[ForeignKeyMeta] = getattr(
            self, "_foreign_key_meta", None
        )
        if foreign_key_meta:
            references = foreign_key_meta.resolved_references
            tablename = references._meta.tablename
            on_delete = foreign_key_meta.on_delete.value
            on_update = foreign_key_meta.on_update.value
            target_column_name = (
                foreign_key_meta.resolved_target_column._meta.name
            )
            query += (
                f" REFERENCES {tablename} ({target_column_name})"
                f" ON DELETE {on_delete}"
                f" ON UPDATE {on_update}"
            )

        # Always ran for Cockroach because unique_rowid() is directly
        # defined for Cockroach Serial and BigSerial.
        # Postgres and SQLite will not run this for Serial and BigSerial.
        if self._meta.engine_type in (
            "cockroach"
        ) or self.__class__.__name__ not in ("Serial", "BigSerial"):
            default = self.get_default_value()
            sql_value = self.get_sql_value(value=default)
            query += f" DEFAULT {sql_value}"

        return query

    def copy(self) -> Column:
        column: Column = copy.copy(self)
        column._meta = self._meta.copy()
        return column

    def __deepcopy__(self, memo) -> Column:
        """
        We override deepcopy, as it's too slow if it has to recreate
        everything.
        """
        return self.copy()

    def __str__(self):
        return self.ddl.__str__()

    def __repr__(self):
        try:
            table = self._meta.table
        except ValueError:
            table_class_name = "Unknown"
        else:
            table_class_name = table.__name__
        return (
            f"{table_class_name}.{self._meta.name} - "
            f"{self.__class__.__name__}"
        )



from unittest import TestCase

from piccolo.table import create_db_tables_sync, drop_db_tables_sync
from tests.base import engines_only, engines_skip
from tests.example_apps.music.tables import Band, Manager


class TestSave(TestCase):
    def setUp(self):
        create_db_tables_sync(Manager, Band)

    def tearDown(self):
        drop_db_tables_sync(Manager, Band)

    def test_save_new(self):
        """
        Make sure that saving a new instance works.
        """
        manager = Manager(name="Maz")

        query = manager.save()
        print(query)
        self.assertTrue("INSERT" in query.__str__())

        query.run_sync()

        names = [i["name"] for i in Manager.select(Manager.name).run_sync()]
        self.assertTrue("Maz" in names)

        manager.name = "Maz2"
        query = manager.save()
        print(query)
        self.assertTrue("UPDATE" in query.__str__())

        query.run_sync()
        names = [i["name"] for i in Manager.select(Manager.name).run_sync()]
        self.assertTrue("Maz2" in names)
        self.assertTrue("Maz" not in names)

    @engines_skip("cockroach")
    def test_save_specific_columns(self):
        """
        Make sure that we can save a subset of columns.
        """
        manager = Manager(name="Guido")
        manager.save().run_sync()

        band = Band(name="Pythonistas", popularity=1000, manager=manager)
        band.save().run_sync()

        self.assertEqual(
            Band.select().run_sync(),
            [
                {
                    "id": 1,
                    "name": "Pythonistas",
                    "manager": 1,
                    "popularity": 1000,
                }
            ],
        )

        band.name = "Pythonistas 2"
        band.popularity = 2000
        band.save(columns=[Band.name]).run_sync()

        # Only the name should update, and not the popularity:
        self.assertEqual(
            Band.select().run_sync(),
            [
                {
                    "id": 1,
                    "name": "Pythonistas 2",
                    "manager": 1,
                    "popularity": 1000,
                }
            ],
        )

        # Also test it using strings to identify columns
        band.name = "Pythonistas 3"
        band.popularity = 3000
        band.save(columns=["popularity"]).run_sync()

        # Only the popularity should update, and not the name:
        self.assertEqual(
            Band.select().run_sync(),
            [
                {
                    "id": 1,
                    "name": "Pythonistas 2",
                    "manager": 1,
                    "popularity": 3000,
                }
            ],
        )

    @engines_only("cockroach")
    def test_save_specific_columns_alt(self):
        """
        Make sure that we can save a subset of columns.
        """
        manager = Manager(name="Guido")
        manager.save().run_sync()

        band = Band(name="Pythonistas", popularity=1000, manager=manager)
        band.save().run_sync()

        self.assertEqual(
            Band.select().run_sync(),
            [
                {
                    "id": band.id,
                    "name": "Pythonistas",
                    "manager": band.manager.id,
                    "popularity": 1000,
                }
            ],
        )

        band.name = "Pythonistas 2"
        band.popularity = 2000
        band.save(columns=[Band.name]).run_sync()

        # Only the name should update, and not the popularity:
        self.assertEqual(
            Band.select().run_sync(),
            [
                {
                    "id": band.id,
                    "name": "Pythonistas 2",
                    "manager": band.manager.id,
                    "popularity": 1000,
                }
            ],
        )

        # Also test it using strings to identify columns
        band.name = "Pythonistas 3"
        band.popularity = 3000
        band.save(columns=["popularity"]).run_sync()

        # Only the popularity should update, and not the name:
        self.assertEqual(
            Band.select().run_sync(),
            [
                {
                    "id": band.id,
                    "name": "Pythonistas 2",
                    "manager": band.manager.id,
                    "popularity": 3000,
                }
            ],
        )



import enum
from unittest import TestCase

from piccolo.columns.column_types import Array, Varchar
from piccolo.table import Table
from tests.base import engines_only
from tests.example_apps.music.tables import Shirt


class TestChoices(TestCase):
    def setUp(self):
        Shirt.create_table().run_sync()

    def tearDown(self):
        Shirt.alter().drop_table().run_sync()

    def _insert_shirts(self):
        Shirt.insert(
            Shirt(size=Shirt.Size.small),
            Shirt(size=Shirt.Size.medium),
            Shirt(size=Shirt.Size.large),
        ).run_sync()

    def test_save(self):
        """
        Make sure saving works, when setting a value as an Enum.
        """
        shirt = Shirt(size=Shirt.Size.large)
        shirt.save().run_sync()

    def test_default(self):
        """
        Make sure the default works correctly, when the default is an Enum.
        """
        Shirt().save().run_sync()
        shirt = Shirt.objects().first().run_sync()
        self.assertEqual(shirt.size, "l")

    def test_update(self):
        """
        Make sure rows can be updated using Enums.
        """
        self._insert_shirts()
        Shirt.update({Shirt.size: Shirt.Size.large}).where(
            Shirt.size == Shirt.Size.small
        ).run_sync()
        shirts = (
            Shirt.select(Shirt.size)
            .output(as_list=True)
            .order_by(Shirt._meta.primary_key)
            .run_sync()
        )
        self.assertEqual(shirts, ["l", "m", "l"])

    def test_select_where(self):
        """
        Make sure Enums can be used in the where clause of select queries.
        """
        self._insert_shirts()
        shirts = (
            Shirt.select(Shirt.size)
            .where(Shirt.size == Shirt.Size.small)
            .run_sync()
        )
        self.assertEqual(shirts, [{"size": "s"}])

    def test_objects_where(self):
        """
        Make sure Enums can be used in the where clause of objects queries.
        """
        self._insert_shirts()
        shirts = (
            Shirt.objects().where(Shirt.size == Shirt.Size.small).run_sync()
        )
        self.assertEqual(len(shirts), 1)
        self.assertEqual(shirts[0].size, "s")


class Ticket(Table):
    class Extras(str, enum.Enum):
        drink = "drink"
        snack = "snack"
        program = "program"

    extras = Array(Varchar(), choices=Extras)


@engines_only("postgres", "sqlite")
class TestArrayChoices(TestCase):
    """
    🐛 Cockroach bug: https://github.com/cockroachdb/cockroach/issues/71908 "could not decorrelate subquery" error under asyncpg
    """  # noqa: E501

    def setUp(self):
        Ticket.create_table().run_sync()

    def tearDown(self):
        Ticket.alter().drop_table().run_sync()

    def test_string(self):
        """
        Make sure strings can be passed in as choices.
        """
        ticket = Ticket(extras=["drink", "snack", "program"])
        ticket.save().run_sync()

        self.assertListEqual(
            Ticket.select(Ticket.extras).run_sync(),
            [{"extras": ["drink", "snack", "program"]}],
        )

    def test_enum(self):
        """
        Make sure enums can be passed in as choices.
        """
        ticket = Ticket(
            extras=[
                Ticket.Extras.drink,
                Ticket.Extras.snack,
                Ticket.Extras.program,
            ]
        )
        ticket.save().run_sync()

        self.assertListEqual(
            Ticket.select(Ticket.extras).run_sync(),
            [{"extras": ["drink", "snack", "program"]}],
        )

    def test_invalid_choices(self):
        """
        Make sure an invalid choices Enum is rejected.
        """
        with self.assertRaises(ValueError) as manager:

            class Ticket(Table):
                # This will be rejected, because the values are ints, and the
                # Array's base_column is Varchar.
                class Extras(int, enum.Enum):
                    drink = 1
                    snack = 2
                    program = 3

                extras = Array(Varchar(), choices=Extras)

        self.assertEqual(
            manager.exception.__str__(), "drink doesn't have the correct type"
        )



from __future__ import annotations

import typing as t

from piccolo.columns.base import Column
from piccolo.query.base import Query
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table


class DropIndex(Query):
    def __init__(
        self,
        table: t.Type[Table],
        columns: t.List[t.Union[Column, str]],
        if_exists: bool = True,
        **kwargs,
    ):
        self.columns = columns
        self.if_exists = if_exists
        super().__init__(table, **kwargs)

    @property
    def column_names(self) -> t.List[str]:
        return [
            i._meta.name if isinstance(i, Column) else i for i in self.columns
        ]

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        column_names = self.column_names
        index_name = self.table._get_index_name(column_names)
        query = "DROP INDEX"
        if self.if_exists:
            query += " IF EXISTS"
        return [QueryString(f"{query} {index_name}")]



from unittest import TestCase

from piccolo.columns import Varchar
from piccolo.table import create_table_class


class TestCreateTableClass(TestCase):
    def test_create_table_class(self):
        """
        Make sure a basic `Table` can be created successfully.
        """
        _Table = create_table_class(class_name="MyTable")
        self.assertEqual(_Table._meta.tablename, "my_table")

        _Table = create_table_class(
            class_name="MyTable", class_kwargs={"tablename": "my_table_1"}
        )
        self.assertEqual(_Table._meta.tablename, "my_table_1")

        column = Varchar()
        _Table = create_table_class(
            class_name="MyTable", class_members={"name": column}
        )
        self.assertIn(column, _Table._meta.columns)

    def test_protected_tablenames(self):
        """
        Make sure that the logic around protected tablenames still works as
        expected.
        """
        with self.assertRaises(ValueError):
            create_table_class(class_name="User")

        with self.assertRaises(ValueError):
            create_table_class(
                class_name="MyUser", class_kwargs={"tablename": "user"}
            )

        # This shouldn't raise an error:
        create_table_class(
            class_name="User", class_kwargs={"tablename": "my_user"}
        )



from __future__ import annotations

import typing as t

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Indexes(Query):
    """
    Returns the indexes for the given table.
    """

    @property
    def postgres_querystrings(self) -> t.Sequence[QueryString]:
        return [
            QueryString(
                "SELECT indexname AS name FROM pg_indexes "
                "WHERE tablename = {}",
                self.table._meta.tablename,
            )
        ]

    @property
    def cockroach_querystrings(self) -> t.Sequence[QueryString]:
        return [
            QueryString(
                "SELECT indexname AS name FROM pg_indexes "
                "WHERE tablename = {}",
                self.table._meta.tablename,
            )
        ]

    @property
    def sqlite_querystrings(self) -> t.Sequence[QueryString]:
        tablename = self.table._meta.tablename
        return [QueryString(f"PRAGMA index_list({tablename})")]

    async def response_handler(self, response):
        return [i["name"] for i in response]



from __future__ import annotations

import typing as t

from piccolo.apps.fixtures.commands.shared import (
    FixtureConfig,
    create_pydantic_fixture_model,
)
from piccolo.conf.apps import Finder
from piccolo.table import sort_table_classes


async def get_dump(
    fixture_configs: t.List[FixtureConfig],
) -> t.Dict[str, t.Any]:
    """
    Gets the data for each table specified and returns a data structure like:

    .. code-block:: python

        {
            'my_app_name': {
                'MyTableName': [
                    {
                        'id': 1,
                        'my_column_name': 'foo'
                    }
                ]
            }
        }

    """
    finder = Finder()

    output: t.Dict[str, t.Any] = {}

    for fixture_config in fixture_configs:
        app_config = finder.get_app_config(app_name=fixture_config.app_name)
        table_classes = [
            i
            for i in app_config.table_classes
            if i.__name__ in fixture_config.table_class_names
        ]
        sorted_table_classes = sort_table_classes(table_classes)

        output[fixture_config.app_name] = {}

        for table_class in sorted_table_classes:
            data = await table_class.select().run()
            output[fixture_config.app_name][table_class.__name__] = data

    return output


async def dump_to_json_string(
    fixture_configs: t.List[FixtureConfig],
) -> str:
    """
    Dumps all of the data for the given tables into a JSON string.
    """
    dump = await get_dump(fixture_configs=fixture_configs)
    pydantic_model = create_pydantic_fixture_model(
        fixture_configs=fixture_configs
    )
    return pydantic_model(**dump).json()


def parse_args(apps: str, tables: str) -> t.List[FixtureConfig]:
    """
    Works out which apps and tables the user is referring to.
    """
    finder = Finder()
    app_names = []

    if apps == "all":
        app_names = finder.get_sorted_app_names()
    elif "," in apps:
        app_names = apps.split(",")
    else:
        # Must be a single app name
        app_names.append(apps)

    table_class_names: t.Optional[t.List[str]] = None

    if tables != "all":
        table_class_names = tables.split(",") if "," in tables else [tables]
    output: t.List[FixtureConfig] = []

    for app_name in app_names:
        app_config = finder.get_app_config(app_name=app_name)
        table_classes = app_config.table_classes

        if table_class_names is None:
            fixture_configs = [i.__name__ for i in table_classes]
        else:
            fixture_configs = [
                i.__name__
                for i in table_classes
                if i.__name__ in table_class_names
            ]
        output.append(
            FixtureConfig(
                app_name=app_name,
                table_class_names=fixture_configs,
            )
        )

    return output


async def dump(apps: str = "all", tables: str = "all"):
    """
    Serialises the data from the given Piccolo apps / tables, and prints it
    out.

    :param apps:
        For all apps, specify `all`. For specific apps, pass in a comma
        separated list e.g. `blog,profiles,billing`. For a single app, just
        pass in the name of that app, e.g. `blog`.
    :param tables:
        For all tables, specify `all`. For specific tables, pass in a comma
        separated list e.g. `Post,Tag`. For a single app, just
        pass in the name of that app, e.g. `Post`.

    """
    fixture_configs = parse_args(apps=apps, tables=tables)
    json_string = await dump_to_json_string(fixture_configs=fixture_configs)
    print(json_string)



from unittest import TestCase

from piccolo.table import (
    create_db_tables_sync,
    drop_db_tables_sync,
    drop_tables,
)
from tests.example_apps.music.tables import Band, Manager


class TestDropTables(TestCase):
    def setUp(self):
        create_db_tables_sync(Band, Manager)

    def test_drop_db_tables(self):
        """
        Make sure the tables are dropped.
        """
        self.assertEqual(Manager.table_exists().run_sync(), True)
        self.assertEqual(Band.table_exists().run_sync(), True)

        drop_db_tables_sync(Manager, Band)

        self.assertEqual(Manager.table_exists().run_sync(), False)
        self.assertEqual(Band.table_exists().run_sync(), False)

    def test_drop_tables(self):
        """
        This is a deprecated function, which just acts as a proxy.
        """
        self.assertEqual(Manager.table_exists().run_sync(), True)
        self.assertEqual(Band.table_exists().run_sync(), True)

        drop_tables(Manager, Band)

        self.assertEqual(Manager.table_exists().run_sync(), False)
        self.assertEqual(Band.table_exists().run_sync(), False)



import typing as t
from unittest import TestCase

from piccolo.columns import Varchar
from piccolo.table import Table
from piccolo.table_reflection import TableStorage
from piccolo.utils.sync import run_sync
from tests.base import engines_only
from tests.example_apps.music.tables import Band, Manager


@engines_only("postgres", "cockroach")
class TestTableStorage(TestCase):
    def setUp(self) -> None:
        self.table_storage = TableStorage()
        for table_class in (Manager, Band):
            table_class.create_table().run_sync()

    def tearDown(self):
        self.table_storage.clear()
        for table_class in (Band, Manager):
            table_class.alter().drop_table(if_exists=True).run_sync()

    def _compare_table_columns(
        self, table_1: t.Type[Table], table_2: t.Type[Table]
    ):
        """
        Make sure that for each column in table_1, there is a corresponding
        column in table_2 of the same type.
        """
        column_names = [
            column._meta.name for column in table_1._meta.non_default_columns
        ]
        for column_name in column_names:
            col_1 = table_1._meta.get_column_by_name(column_name)
            col_2 = table_2._meta.get_column_by_name(column_name)

            # Make sure they're the same type
            self.assertEqual(type(col_1), type(col_2))

            # Make sure they're both nullable or not
            self.assertEqual(col_1._meta.null, col_2._meta.null)

            # Make sure the max length is the same
            if isinstance(col_1, Varchar) and isinstance(col_2, Varchar):
                self.assertEqual(col_1.length, col_2.length)

            # Make sure the unique constraint is the same
            self.assertEqual(col_1._meta.unique, col_2._meta.unique)

    def test_reflect_all_tables(self):
        run_sync(self.table_storage.reflect())
        reflected_tables = self.table_storage.tables
        self.assertEqual(len(reflected_tables), 2)
        for table_class in (Manager, Band):
            self._compare_table_columns(
                reflected_tables[table_class._meta.tablename], table_class
            )

    def test_reflect_with_include(self):
        run_sync(self.table_storage.reflect(include=["manager"]))
        reflected_tables = self.table_storage.tables
        self.assertEqual(len(reflected_tables), 1)
        self._compare_table_columns(reflected_tables["manager"], Manager)

    def test_reflect_with_exclude(self):
        run_sync(self.table_storage.reflect(exclude=["band"]))
        reflected_tables = self.table_storage.tables
        self.assertEqual(len(reflected_tables), 1)
        self._compare_table_columns(reflected_tables["manager"], Manager)

    def test_get_present_table(self):
        run_sync(self.table_storage.reflect())
        table = run_sync(self.table_storage.get_table(tablename="manager"))
        self._compare_table_columns(table, Manager)

    def test_get_unavailable_table(self):
        run_sync(self.table_storage.reflect(exclude=["band"]))
        # make sure only one table is present
        self.assertEqual(len(self.table_storage.tables), 1)
        table = run_sync(self.table_storage.get_table(tablename="band"))
        # make sure the returned table is correct
        self._compare_table_columns(table, Band)
        # make sure the requested table has been added to the TableStorage
        self.assertEqual(len(self.table_storage.tables), 2)
        self.assertIsNotNone(self.table_storage.tables.get("band"))

    def test_get_schema_and_table_name(self):
        tableNameDetail = self.table_storage._get_schema_and_table_name(
            "music.manager"
        )
        self.assertEqual(tableNameDetail.name, "manager")
        self.assertEqual(tableNameDetail.schema, "music")
